
from . import _VDAPIService, _VDDuplicateableResponse

class _DemandTagKeyAPI(_VDAPIService):

    __API__ = "demand_tags"

    def __init__(self, demand_tag):
        super(_DemandTagKeyAPI, self).__init__()
        self.demand_tag_id = demand_tag.id
        self.account_id = demand_tag.account_id

    @property
    def endpoint(self):
        """
        The api endpoint that is used for this service.  For example:: 
            
            In [1]: import springserve

            In [2]: springserve.supply_tags.endpoint
            Out[2]: '/supply_tags'

        """
        return "/demand_tags/{}/demand_tag_keys".format(self.demand_tag_id)

class _DemandTagResponse(_VDDuplicateableResponse):
    
    def get_key_value_targeting_keys(self):
        # Need to make a new one per bill
        return _DemandTagKeyAPI(self).get()

    def add_key(self, data, **kwargs):
        return _DemandTagKeyAPI(self).post(data, **kwargs)

class _DemandTagAPI(_VDAPIService):

    __RESPONSE_OBJECT__ = _DemandTagResponse
    __API__ = "demand_tags"

class _DemandPartnerAPI(_VDAPIService):

    __API__ = "demand_partners"

class _DemandGroupAPI(_VDAPIService):

    __API__ = "demand_groups"









from . import _VDAPIService, _VDDuplicateableResponse

class _SupplyTagAPI(_VDAPIService):

    __RESPONSE_OBJECT__ = _VDDuplicateableResponse
    __API__ = "supply_tags"


class _SupplyPartnerAPI(_VDAPIService):

    __API__ = "supply_partners"

class _SupplyGroupAPI(_VDAPIService):

    __API__ = "supply_groups"







from . import _VDAPIService, _VDAPIResponse, _VDAPIMultiResponse

from datetime import datetime

from time import sleep

import pandas

class _ReportingResponse(_VDAPIMultiResponse):

    def __init__(self, service, api_response_data, path_params, query_params,
                ok, payload):

        super(_ReportingResponse, self).__init__(service, api_response_data,
                                                path_params, query_params, _ReportingResponse,
                                                ok, payload)
        self.dataframe = None

    def _build_cache(self, objects):
        pass

    def to_dataframe(self):
        if self.dataframe is None:
            self.dataframe = pandas.DataFrame(self.raw['data'])
        return self.dataframe

    def _is_last_page(self, resp):
        self._all_pages_gotten = not resp.data
        return self._all_pages_gotten

    def get_next_page(self, clear_previous=True):

        if self._all_pages_gotten:
            return False

        self._payload['page'] = self._current_page+1
        if self._payload:
            resp = self._service.post(data=self._payload)
        else:
            raise('Original report paramaters are missing')

        if self._is_last_page(resp):
            self._all_pages_gotten = True
            return False

        new_data = pandas.DataFrame(resp.raw['data'])
        #clear previous data
        if clear_previous:
            self.dataframe = new_data
        else:
            #call to dataframe to make sure self.dataframe is set
            self.to_dataframe()
            self.dataframe = self.dataframe.append(new_data)
        self._current_page += 1
        return True

    def get_all_pages(self):
        while not self._all_pages_gotten:
            self.get_next_page(clear_previous=False)
        return None

class _ReportingAPI(_VDAPIService):

    __API__ = "report"
    __RESPONSES_OBJECT__ = _ReportingResponse

    INTERVALS = ("hour", "day", "cumulative")

    def _format_date(self, date):

        if isinstance(date, datetime):
            return date.strftime("%Y-%m-%d")
        return date

    def _get_report(self, payload):
        response = self.post(data=payload)
        #something bad happened
        if not response.ok:
            return response
        if 'report_id' not in response.raw:
            raise('report_id field not in response: {}'.format(response.raw))
        self._report_id = response.raw['report_id']
        payload['report_id'] = self._report_id
        if 'status' not in response.raw:
            raise('status field not in response: {}'.format(response.raw))

        #poll the api untill we get a completed report
        while response.raw['status'] != 'COMPLETE':
            sleep(1)
            response = self.post(data=payload)
        return response

    def build_response(self, api_response, path_params, query_params, payload):
        is_ok = api_response.ok

        if not is_ok and api_response.status_code == 401:
            raise VDAuthError("Need to Re-Auth")
        if api_response.status_code == 204:  # this means empty
            resp_json = {}
        else:
            resp_json = api_response.json

        return self.__RESPONSES_OBJECT__(self, resp_json, path_params, query_params, is_ok, payload)


    def run(self, start_date=None, end_date=None, interval=None, dimensions=None,
            account_id=None, **kwargs):
        """
        parameter     options (if applicable)  notes
        ===================================================
        start_date:  "2015-12-01 00:00:00" or "2015-12-01"
        end_date:    "2015-12-02 00:00:00" or "2015-12-01"
        interval:    "hour", "day", "cumulative"
        timezone:    "UTC", "America/New_York"   defaults to America/New_York
        date_range:  Today, Yesterday, Last 7 Days   date_range takes precedence over start_date/end_date
        dimensions:  supply_tag_id, demand_tag_id, detected_domain, declared_domain, demand_type, supply_type, supply_partner_id, demand_partner_id, supply_group  domain is only available when using date_range of Today, Yesterday, or Last 7 Days

        the following parameters act as filters; pass an array of values (usually IDs)
        =================================================================================

        supply_tag_ids:  [22423,22375, 25463]
        demand_tag_ids:  [22423,22375, 25463]
        detected_domains:         ["nytimes.com", "weather.com"]
        declared_domains:         ["nytimes.com", "weather.com"]
        supply_types     ["Syndicated","Third-Party"]
        supply_partner_ids:  [30,42,41]
        supply_group_ids:    [13,15,81]
        demand_partner_ids:  [3,10,81]
        demand_types:    ["Vast Only","FLASH"]
        """
        self.payload = {
            'start_date': self._format_date(start_date),
            'end_date': self._format_date(end_date),
            'report_service': True,
            'async': True
        }

        if interval:
            if interval not in self.INTERVALS:
                raise Exception("not a valid interval")
            self.payload['interval'] = interval

        if dimensions:
            self.payload['dimensions'] = dimensions

        if account_id:
            self.payload['account_id'] = account_id

        if kwargs:
            self.payload.update(kwargs)

        return self._get_report(self.payload)


class _TrafficQualityReport(_ReportingAPI):

    __API__ = "traffic_quality_reports"
    __RESPONSES_OBJECT__ = _ReportingResponse





"""
AWS ELB 5XX error looks like:

<html>
<head><title>503 Service Temporarily Unavailable</title></head>
<body bgcolor="white">
<center><h1>503 Service Temporarily Unavailable</h1></center>
</body>
</html>
"""
import time

_msg = None
try:
    from link import lnk
    from link.wrappers.springservewrappers import SpringServeAPIResponseWrapper
    _msg = lnk.msg
except:
    pass


AWS_ELB_ERROR_MESSAGES = ("503 Service Temporarily Unavailable", "502 Bad Gateway")
RACK_ATTACK_STATUS_CODE = 429
RACK_ATTACK_MESSAGE = "Retry later\n"


def raw_response_retry(api_call, limit=4, sleep_duration=5, backoff_factor=2):
    """
    Decorator for SpringServe API to handle retries (with exponential backoff) in the case
    of a rate-limit or 5XX error.

    Sleep duration and backoff factor control wait time between successive failures, e.g.
    sleep_duration 3 and backoff_factor 2 means sleep 3s, 6s, 12s, 24s

    :param int limit: Max number of retry attempts
    :param int sleep_duration: Initial sleep time
    :param float/int backoff_factor: Factor to increase sleep between successive retries.
    """
    def wrapped(*args, **kwargs):
        sleeps = sleep_duration
        num_attempts = 0
        while num_attempts < limit:
            # make the API call
            resp = api_call(*args, **kwargs)

            aws_check = (
                    # make sure it's the link response object
                    isinstance(resp, SpringServeAPIResponseWrapper) and
                    # HTTP status codes that are valid for retries
                    resp.status_code >= 500 and resp.status_code < 600 and
                    # content matches one of our error messages - note that ELB error
                    # messages will not be JSON (they are HTML strings) so cannot check
                    # resp.json attribute, as this will not always be valid
                    any([e in resp.content for e in AWS_ELB_ERROR_MESSAGES])
                    )

            rack_attack_check = (
                    isinstance(resp, SpringServeAPIResponseWrapper) and
                    resp.status_code == RACK_ATTACK_STATUS_CODE and
                    resp.content == RACK_ATTACK_MESSAGE
                    )

            if aws_check or rack_attack_check:
                _msg.warn("Encountered rate-limit (attempt {}), sleeping".format(num_attempts))
                num_attempts += 1
                time.sleep(sleeps)
                sleeps *= backoff_factor
            # call was either successful, or an error outside of the purview of this
            # handler
            else:
                return resp

        # We've hit max retry attempts, return anyways
        return resp

    return wrapped
